/* sslio.c    - SSL io layer interface
 *
 * This provides a layer implementation that uses a BIO that uses the lower
 * layer to do the I/O work
 *
 * http://mozilla-crypto.ssleay.org
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "moz_eay.h"

/* For the PRFileSec
 * method     == ssleay_ssl_fd_method
 * secret     == (PRFilePrivate *)SSL
 * PRFileDesc == ssleay_ssl_destructor
 * identity   == "SSL (SSLeay)"
 */
#define SSLEAY_SSL_IDENTITY	"SSL (SSLeay)"

/* return PR_INVALID_METHOD_ERROR if not implemented
 * fd->lower->methods->close(fd->lower)) to call next one down
 */

static PRInt32 ssleay_ssl_send(PRFileDesc *fd, const void *buf, 
		PRInt32 amount, PRIntn flags, PRIntervalTime timeout);

void ssleay_ssl_destructor(prfd)
PRFileDesc *prfd;
	{
	SSL *ssl;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_destructor\n");

	ssl=(SSL *)prfd->secret;
	if (ssl != NULL)
		SSL_free(ssl);
	}

static PRStatus ssleay_ssl_close(prfd)
PRFileDesc *prfd;
	{
	PRStatus ret;
	SSL *ssl=(SSL *)prfd->secret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_close\n");

	if ((ret=prfd->lower->methods->close(prfd->lower)) == PR_SUCCESS)
		{
		SSL_free(ssl);
		prfd->secret=NULL;
		}
	return(ret);
	}

static PRStatus ssleay_ssl_read(prfd,buf,amount)
PRFileDesc *prfd;
void *buf;
PRInt32 amount;
	{
	PRStatus ret;
	SSL *ssl=(SSL *)prfd->secret;
	BIO *bio;


	bio=SSL_get_rbio(ssl);
	BIO_ctrl(bio,BIO_C_SET_PR_FILEDESC,0,(char *)prfd->lower);
	ret=SSL_read(ssl,(unsigned char *)buf,(int)amount);
	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_read -> %d %c\n",ret,
			BIO_should_read(bio)?'R':(BIO_should_write(bio)?'W':'X'));
	return(ret);
	}

static PRStatus ssleay_ssl_write(prfd,buf,amount)
PRFileDesc *prfd;
void *buf;
PRInt32 amount;
	{
	PRStatus ret;
	SSL *ssl=(SSL *)prfd->secret;
	BIO *bio;


	bio=SSL_get_rbio(ssl);
	BIO_ctrl(bio,BIO_C_SET_PR_FILEDESC,0,(char *)prfd->lower);
	ret=SSL_write(ssl,(unsigned char *)buf,(int)amount);
	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_write %d %c\n",ret,
			BIO_should_read(bio)?'R':(BIO_should_write(bio)?'W':'X'));
	return(ret);
	}

static PRInt32 ssleay_ssl_pending(prfd)
PRFileDesc *prfd;
	{
	PRInt32 ret;
	SSL *ssl=(SSL *)prfd->secret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_pending\n");
	ret=SSL_pending(ssl);
	return(ret);
	}

static PRInt64 ssleay_ssl_pending64(prfd)
PRFileDesc *prfd;
	{
	PRInt64 ret;
	SSL *ssl=(SSL *)prfd->secret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_pending64\n");
	ret=SSL_pending(ssl);
	return(ret);
	}

static PRStatus ssleay_ssl_sync(prfd)
PRFileDesc *prfd;
	{
	PRStatus ret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_sync\n");
	ret=prfd->lower->methods->fsync(prfd->lower);
	return(ret);
	}

static PRInt32 ssleay_ssl_seek(prfd,offset,whence)
PRFileDesc *prfd;
PRInt32 offset;
PRSeekWhence whence;
	{
	PRInt32 ret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_seek\n");
	ret=prfd->lower->methods->seek(prfd->lower,offset,whence);
	return(ret);
	}

static PRInt64 ssleay_ssl_seek64(prfd,offset,whence)
PRFileDesc *prfd;
PRInt64 offset;
PRSeekWhence whence;
	{
	PRInt64 ret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_seek64\n");
	ret=prfd->lower->methods->seek(prfd->lower,offset,whence);
	return(ret);
	}

static PRStatus ssleay_ssl_not_implemented()
	{
	return(PR_INVALID_METHOD_ERROR);
	}

static PRInt32 ssleay_ssl_writev(prfd,iov,size,timeout)
PRFileDesc *prfd;
PRIOVec *iov;
PRInt32 size;
PRIntervalTime timeout;
	{
	PRInt32 ret=0,tmp;
	int i,to;
	PRIntervalTime start,end;
	char *p;
	int num;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_writev\n");

	to=(timeout != PR_INTERVAL_NO_TIMEOUT);

	for (i=0; i<size; i++)
		{
		num=iov->iov_len;
		p=iov->iov_base;
		while (num > 0)
			{
			if (to)
				start=PR_IntervalNow();
			tmp=ssleay_ssl_send(prfd,p,num,0,timeout);
			if (tmp <= 0) return(tmp);
			num-=tmp;
			p+=tmp;
			ret+=tmp;
			if (to)
				{
				end=PR_IntervalNow();
				if ((end-start) >= timeout)
					{
					PR_SetError(PR_IO_TIMEOUT_ERROR,0);
					return(-1);
					}
				else
					timeout-=(end-start);
				}
			}
		}
	return(tmp);
	}

static PRStatus ssleay_ssl_connect(prfd,addr,timeout)
PRFileDesc *prfd;
PRNetAddr *addr;
PRIntervalTime timeout;
	{
	PRStatus ret;
	SSL *ssl=(SSL *)prfd->secret;


	ret=prfd->lower->methods->connect(prfd->lower,addr,timeout);
	if (ret == PR_SUCCESS)
		{
		SSL_clear(ssl);
		SSL_set_connect_state(ssl);
		}
	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_connect -> %d\n",ret);
	return(ret);
	}

/* This one needs work, but I don't think it really matters at this
 * point in time.
 */
static PRFileDesc *ssleay_ssl_accept(prfd,addr,timeout)
PRFileDesc *prfd;
PRNetAddr *addr;
PRIntervalTime timeout;
	{
	PRFileDesc *ret;
	SSL *ssl=(SSL *)prfd->secret;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_accept\n");

	ret=prfd->lower->methods->accept(prfd->lower,addr,timeout);
	if (ret != NULL)
		{
		SSL_clear(ssl);
		SSL_set_accept_state(ssl);
		}
	return(ret);
	}

static PRStatus ssleay_ssl_bind(prfd,addr)
PRFileDesc *prfd;
PRNetAddr *addr;
	{
	PRStatus ret;

	ret=prfd->lower->methods->bind(prfd->lower,addr);
	return(ret);
	}

static PRStatus ssleay_ssl_listen(prfd,backlog)
PRFileDesc *prfd;
PRIntn backlog;
	{
	PRStatus ret;

	ret=prfd->lower->methods->listen(prfd->lower,backlog);
	return(ret);
	}

static PRStatus ssleay_ssl_shutdown(prfd,how)
PRFileDesc *prfd;
PRShutdownHow how;
	{
	PRStatus ret;
	SSL *ssl=(SSL *)prfd->secret;

	SSL_shutdown(ssl);
	ret=prfd->lower->methods->shutdown(prfd->lower,how);
	return(ret);
	}

static PRInt32 ssleay_ssl_recv(prfd,buf,amount,flags,timeout)
PRFileDesc *prfd;
void *buf;
PRInt32 amount;
PRIntn flags;
PRIntervalTime timeout;
	{
	PRInt32 ret;
	SSL *ssl=(SSL *)prfd->secret;
	BIO *bio;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_recv\n");

	bio=SSL_get_rbio(ssl);
	BIO_ctrl(bio,BIO_C_SET_PR_FILEDESC,0,(char *)prfd->lower);
	BIO_ctrl(bio,BIO_C_SET_PR_TIMEOUT,timeout,NULL);

	ret=SSL_read(ssl,(unsigned char *)buf,(int)amount);
	return(ret);
	}

static PRInt32 ssleay_ssl_send(prfd,buf,amount,flags,timeout)
PRFileDesc *prfd;
const void *buf;
PRInt32 amount;
PRIntn flags;
PRIntervalTime timeout;
	{
	PRInt32 ret;
	SSL *ssl=(SSL *)prfd->secret;
	BIO *bio;

	if (dbglev)
		BIO_printf(bio_err,"sslio.c: ssleay_ssl_send\n");

	bio=SSL_get_wbio(ssl);
	BIO_ctrl(bio,BIO_C_SET_PR_FILEDESC,0,(char *)prfd->lower);
	BIO_ctrl(bio,BIO_C_SET_PR_TIMEOUT,timeout,NULL);

	ret=SSL_write(ssl,(unsigned char *)buf,(int)amount);
	return(ret);
	}

static PRInt32 ssleay_ssl_recvfrom(prfd,buf,amount,flags,addr,timeout)
PRFileDesc *prfd;
void *buf;
PRInt32 amount;
PRIntn flags;
PRNetAddr *addr;
PRIntervalTime timeout;
	{
	return(PR_INVALID_METHOD_ERROR);
	}

static PRInt32 ssleay_ssl_sendto(prfd,buf,amount,flags,addr,timeout)
PRFileDesc *prfd;
void *buf;
PRInt32 amount;
PRIntn flags;
PRNetAddr *addr;
PRIntervalTime timeout;
	{
	return(PR_INVALID_METHOD_ERROR);
	}


/*
	nsprpub/pr/src/io/prio.c:184
	What this means is that we want a read poll handler to check for
	pending data.
*/

static PRInt16 ssleay_ssl_poll(prfd,how_flags)
PRFileDesc *prfd;
PRInt16 how_flags;
	{
	PRInt32 ret;
	PRInt16 new;
	SSL *ssl=(SSL *)prfd->secret;

	if ((how_flags & PR_POLL_READ) &&
		SSL_pending(ssl))
		return(PR_POLL_READ);
	else
		return(0);
	}

static PRInt32 ssleay_ssl_transmitfile(prnfd,prffd,headers,hlen,flags,timeout)
PRFileDesc *prnfd;
PRFileDesc *prffd;
const void *headers;
PRInt32 hlen;
PRTransmitFileFlags flags;
PRIntervalTime timeout;
	{
#if 0
	PRInt32 ret;

	/* This one needs more work .... */

	ret=_PR_UnixTransmitFile(prnfd,prffd,headers,hlen,flags,timeout);
	return(ret);
#else
	return(PR_INVALID_METHOD_ERROR);
#endif
	}

#if 0
/* We need to push on the SSL envionment... this call is almost definitly
 * not needed but is an artifact of the Netscape server.... */
/* This function is heavilly plundered from else where :-) - eay */
static PRInt32 ssleay_ssl_acceptread(lprfd,aprfd,peeraddr,buf,amount,timeout)
PRFileDesc *lprfd;
PRFileDesc **aprfd;
PRNetAddr **peeraddr,
void *buf;
PRInt32 amount;
PRIntervalTime timeout;
	{
	PRInt32 rv = -1;
	PRNetAddr remote;
	PRFileDesc *accepted = NULL;
	PRIntervalTime start, elapsed;

	if (pt_TestAbort()) return rv;

	if (PR_INTERVAL_NO_TIMEOUT != timeout) start = PR_IntervalNow();
	if ((accepted = PR_Accept(lprfd->lower, &remote, timeout)) == NULL) return rv;

	if (PR_INTERVAL_NO_TIMEOUT != timeout)
		{
		elapsed = (PRIntervalTime) (PR_IntervalNow() - start);
		if (elapsed > timeout)
			{
			PR_SetError(PR_IO_TIMEOUT_ERROR, 0);
			goto failed;
			}
		else timeout = timeout - elapsed;
		}

	rv = PR_Recv(accepted, buf, amount, 0, timeout);
	if (rv >= 0)
		{
		*aprfd = accepted;
		*raddr = (PRNetAddr *)((char*)buf + amount);
		memcpy(*raddr, &remote, PR_NETADDR_SIZE(&remote));
		return rv;
		}
failed:
	PR_Close(accepted);
	return rv;
	}
#endif

static PRStatus ssleay_ssl_getsockname(prfd,addr)
PRFileDesc *prfd;
PRNetAddr *addr;
	{
	PRStatus ret;

	ret=prfd->lower->methods->getsockname(prfd,addr);
	return(ret);
	}

static PRStatus ssleay_ssl_getpeername(prfd,addr)
PRFileDesc *prfd;
PRNetAddr *addr;
	{
	PRStatus ret;

	ret=prfd->lower->methods->getpeername(prfd,addr);
	return(ret);
	}

static PRStatus ssleay_ssl_getsockopt(prfd,opt,data,optlen)
PRFileDesc *prfd;
PRSockOption opt;
PRSocketOptionData *data;
PRInt32 *optlen;
	{
	PRStatus ret;

	ret=prfd->lower->methods->getsockopt(prfd,opt,data,optlen);
	return(ret);
	}

static PRStatus ssleay_ssl_setsockopt(prfd,optname,data,optlen)
PRFileDesc *prfd;
PRSockOption optname;
PRSocketOptionData *data;
PRInt32 optlen;
	{
	PRStatus ret;

	ret=prfd->lower->methods->setsockopt(prfd,optname,data,optlen);
	return(ret);
	}

static long bio_dump_cb(bio,cmd,argp,argi,argl,ret)
BIO *bio;
int cmd;
char *argp;
int argi;
long argl;
long ret;
        {
        BIO *out;
 
        out=(BIO *)BIO_get_callback_arg(bio);
        if (out == NULL) return(ret);
 
        if (cmd == (BIO_CB_READ|BIO_CB_RETURN))
                {
                BIO_printf(out,"read from %08X [%08lX] (%d bytes => %ld (0x%X))\n",
                        bio,argp,argi,ret,ret);
                BIO_dump(out,argp,(int)ret);
                return(ret);
                }
        else if (cmd == (BIO_CB_WRITE|BIO_CB_RETURN))
                {
                BIO_printf(out,"write to %08X [%08lX] (%d bytes => %ld (0x%X))\n",
                        bio,argp,argi,ret,ret);
                BIO_dump(out,argp,(int)ret);
                }
        return(ret);
        }
 
static PRIOMethods ssleay_ssl_fd_method=
	{
	PR_DESC_LAYERED,		/* PRDescType file_type; */
	ssleay_ssl_close,		/* PRCloseFN close; */
	ssleay_ssl_read,		/* PRReadFN read; */
	ssleay_ssl_write,		/* PRWriteFN write; */
	ssleay_ssl_pending,		/* PRAvailableFN available; */
	ssleay_ssl_pending64,		/* PRAvailable64FN available64; */
	ssleay_ssl_sync,		/* PRFsyncFN fsync; */
	ssleay_ssl_seek,		/* PRSeekFN seek; */
	ssleay_ssl_seek64,		/* PRSeek64FN seek64; */
	(PRFileInfoFN)ssleay_ssl_not_implemented,/* PRFileInfoFN fileInfo; */
	(PRFileInfo64FN)ssleay_ssl_not_implemented,/* PRFileInfo64FN fileInfo64; */
	ssleay_ssl_writev,/* PRWritevFN writev; */
	ssleay_ssl_connect,		/* PRConnectFN connect; */
	ssleay_ssl_accept,		/* PRAcceptFN accept; */
	ssleay_ssl_bind,		/* PRBindFN bind; */
	ssleay_ssl_listen,		/* PRListenFN listen; */
	ssleay_ssl_shutdown,		/* PRShutdownFN shutdown; */
	ssleay_ssl_recv,		/* PRRecvFN recv; */
	ssleay_ssl_send,		/* PRSendFN send; */
	ssleay_ssl_recvfrom,		/* PRRecvfromFN recvfrom; */
	ssleay_ssl_sendto,		/* PRSendtoFN sendto; */
	ssleay_ssl_poll,		/* PRPollFN poll; */
	(PRAcceptreadFN)ssleay_ssl_not_implemented,/* PRAcceptreadFN acceptread; */
	ssleay_ssl_transmitfile,	/* PRTransmitfileFN transmitfile; */
	ssleay_ssl_getsockname,		/* PRGetsocknameFN getsockname; */
	ssleay_ssl_getpeername,		/* PRGetpeernameFN getpeername; */
	ssleay_ssl_getsockopt,		/* PRGetsockoptFN getsockopt; */
	ssleay_ssl_setsockopt,		/* PRSetsockoptFN setsockopt; */
	};

void do_popup(ssl)
SSL *ssl;
	{
	MWContext *context;
	BIO *bio;
	BUF_MEM *bm;
	STACK *sk;
	X509 *peer;
	SSL_CIPHER *c;
	char *str;
	int i;
	unsigned char buf[512];

	bio=BIO_new(BIO_s_mem());
	if (bio == NULL) return;

	context=(MWContext *)SSL_get_ex_data(ssl,ssl_window_id_idx);

	BIO_printf(bio,"Cryptozilla - SSL demo screen\n\n");
	BIO_printf(bio,"Mozilla Crypto Group - http://mozilla-crypto.ssleay.org \n\n");

	/*
	BIO_printf(bio,"Unable to find url, I don't know the mozilla internals well enough yet :-)\n\n");
	*/

	sk=SSL_get_peer_cert_chain(ssl);
	if (sk != NULL)
		{
		BIO_printf(bio,"Certificate chain\n");
		for (i=0; i<sk_num(sk); i++)
			{
			X509_NAME_oneline(X509_get_subject_name((X509 *)
				sk_value(sk,i)),buf,511);
			BIO_printf(bio,"%2d s:%s\n",i,buf);
			X509_NAME_oneline(X509_get_issuer_name((X509 *)
				sk_value(sk,i)),buf,511);
			BIO_printf(bio,"   i:%s\n",buf);
			}
		BIO_printf(bio,"---\n");
		}
	peer=SSL_get_peer_certificate(ssl);
	if (peer != NULL)
		{
		BIO_printf(bio,"Server certificate\n");
		X509_NAME_oneline(X509_get_subject_name(peer),
			buf,511);
		BIO_printf(bio,"subject=%s\n",buf);
		X509_NAME_oneline(X509_get_issuer_name(peer),
			buf,511);
		BIO_printf(bio,"issuer=%s\n",buf);
		}
	BIO_printf(bio,(SSL_session_reused(ssl))?
		"---\nReused, ":"---\nNew, ");
	c=SSL_get_current_cipher(ssl);
	BIO_printf(bio,"%s, Cipher is %s\n",
		SSL_CIPHER_get_version(c),
		SSL_CIPHER_get_name(c));
	if (peer != NULL)
		BIO_printf(bio,"Server public key is %d bit\n",
			EVP_PKEY_bits(X509_get_pubkey(peer)));
	SSL_SESSION_print(bio,SSL_get_session(ssl));
	if (peer != NULL) X509_free(peer);

	BIO_printf(bio,"\nThis software is not for use inside the USA\n");
	BIO_printf(bio,"since it may violate RSA Data Security's\n");
	BIO_printf(bio,"software patent on the RSA encryption algorithm\n");

        /*
	BIO_printf(bio,"\nAnd yes, this is the wrong dialog, but this\n");
	BIO_printf(bio,"is a quick hack to give security information feedback\n");
	*/

	i=BIO_get_mem_ptr(bio,&bm);
	/* There should be space :-) */
	bm->data[bm->length]='\0';

	if (bio_err != NULL)
		BIO_write(bio_err,bm->data,bm->length);
	FE_Confirm(context,bm->data);
	BIO_free(bio);
	}

/* For the PRFileSec
 * method     == ssleay_ssl_fd_method
 * secret     == (PRFilePrivate *)SSL
 * PRFileDesc == ssleay_ssl_destructor
 * identity   == "SSL (SSLeay)"
 */

static void PR_CALLBACK ssleay_ssl_state_callback(s,where,ret)
SSL *s;
int where;
int ret;
	{
	char *str;
	int w;

	w=where& ~SSL_ST_MASK;
	if (ret < 0) return;

	if (w & SSL_ST_CONNECT) str="SSL_connect";
	else if (w & SSL_ST_ACCEPT) str="SSL_accept";
	else str="undefined";

	if ((where & SSL_CB_LOOP) && (bio_err != NULL))
		{
		if (dbglev)
		  BIO_printf(bio_err,"%s:%s\n",str,SSL_state_string_long(s));
		}
	else if ((where & SSL_CB_ALERT) && (bio_err != NULL))
		{
		str=(where & SSL_CB_READ)?"read":"write";
		if (dbglev)
		  BIO_printf(bio_err,"SSL3 alert %s:%s:%s\n",
			str,
			SSL_alert_type_string_long(ret),
			SSL_alert_desc_string_long(ret));
		}
	else if (where & SSL_CB_EXIT)
		{
		if ((ret == 0) && (bio_err != NULL))
			{
			if (dbglev)
			  BIO_printf(bio_err,"%s:failed in %s\n",
				str,SSL_state_string_long(s));
			}
		else if ((ret < 0) && (bio_err != NULL))
			{
			if (dbglev)
			  BIO_printf(bio_err,"%s:error in %s\n",
				str,SSL_state_string_long(s));
			}
		else
			{ /* do a popup */
			do_popup(s);
			}
		}
	}

PRFileDesc *ssl_SSLeay_new(window)
MWContext *window;
	{
	PRFileDesc *ret;
	BIO *bio;
	SSL *ssl;

	ret=PR_CreateIOLayerStub(ssleay_ssl_identity,&ssleay_ssl_fd_method);

	ssl=SSL_new(ssl_ctx);
	SSL_set_ex_data(ssl,ssl_window_id_idx,(char *)window);

	SSL_set_info_callback(ssl,ssleay_ssl_state_callback);

	SSL_set_connect_state(ssl);
	ret->secret=(PRFilePrivate *)ssl;

	bio=BIO_new(BIO_f_mozilla());

#if 0
        /* if we really want to generate a full trace of the I/O then
	 * we could enable this ... 
	 *
	 * TODO - make this runtime configurable
	 */
	if (dbglev) {
	  BIO_set_callback(bio,bio_dump_cb);
	  BIO_set_callback_arg(bio,bio_err);
	}
#endif

	SSL_set_bio(ssl,bio,bio);

	return(ret);
	}

